@@ -17,15 +17,25 @@ except ImportError:
     # We're running in a tree that doesn't have a _version.py, so we don't know what our version is.
     __version__ = "0.0.0"
 
+# We must dynamically find libjvm.so since its unlikely to be in the same place
+# as it was on the distribution on which javabridge was built.
+import sys
+if sys.platform.startswith('linux'):
+    from .locate import find_jre_bin_jdk_so
+    _, jdk_so = find_jre_bin_jdk_so()
+    if jdk_so:
+        import ctypes
+        ctypes.cdll.LoadLibrary(jdk_so)
+
 _jars_dir = os.path.join(os.path.dirname(__file__), 'jars')
 
 #: List of absolute paths to JAR files that are required for the
 #: Javabridge to work.
 JARS = [os.path.realpath(os.path.join(_jars_dir, name + '.jar'))
-        for name in ['rhino-1.7R4', 'runnablequeue']]
+        for name in ['rhino-1.7R4', 'runnablequeue', 'cpython']]
 
 
-from .jutil import start_vm, kill_vm, activate_awt, deactivate_awt
+from .jutil import start_vm, kill_vm, vm, activate_awt, deactivate_awt
 
 from .jutil import attach, detach, get_env
 
@@ -41,7 +51,7 @@ from .jutil import call, get_static_field, static_call, \
 
 # Make Python object that wraps a Java object
 from .jutil import make_method, make_new, make_call, box
-from .wrappers import JWrapper, JClassWrapper
+from .wrappers import JWrapper, JClassWrapper, JProxy
 
 from .jutil import get_nice_arg
 
@@ -66,10 +76,16 @@ from .jutil import JavaError, JavaException, JVMNotFoundError
 
 from ._javabridge import mac_enter_run_loop, mac_stop_run_loop, mac_run_loop_init
     
-# Don't expose: AtExit, attach_ext_env, get_nice_args,
+# References
+from .jutil import create_jref, redeem_jref, create_and_lock_jref,\
+     lock_jref, unlock_jref
+
+# Don't expose: AtExit, get_nice_args,
 # make_run_dictionary, run_in_main_thread, split_sig, unwrap_javascript,
 # print_all_stack_traces
 
 
 # Low-level API
 from ._javabridge import JB_Env, JB_Object, JB_Class
+# JNI helpers.
+from ._javabridge import jni_enter, jni_exit, jvm_enter
