@@ -9,8 +9,9 @@ Copyright (c) 2009-2013 Broad Institute
 All rights reserved.
 
 '''
+from __future__ import print_function
+
 
-import codecs
 import gc
 import inspect
 import logging
@@ -24,10 +25,22 @@ import sys
 import uuid
 from .locate import find_javahome
 import javabridge
+import weakref
+
+
+# long and int are the same type in Py3
+if sys.version_info[0] >= 3:
+    long = int
 
 logger = logging.getLogger(__name__)
 
 
+if sys.version_info >= (3, 0, 0):
+    # basestring -> str and unicode -> str in Python 3
+    basestring = str
+    unicode = str
+
+
 class JavaError(ValueError):
     '''An error caused by using the Javabridge incorrectly'''
     def __init__(self, message=None):
@@ -79,9 +92,11 @@ def _find_jvm_windows():
                 if os.path.isfile(os.path.join(jvm_dir, "jvm.dll")):
                     new_path = ';'.join((os.environ['PATH'], jvm_dir, jre_bin))
                     if isinstance(os.environ['PATH'], str) and \
-                       isinstance(new_path, unicode):
+                       isinstance(new_path, unicode) and \
+                       sys.version_info < (3, 0, 0):
                         # Don't inadvertantly set an environment variable
                         # to unicode: causes subprocess.check_call to fail
+                        # in Python 2
                         new_path = new_path.encode("utf-8")
                     os.environ['PATH'] = new_path
                     found_jvm = True
@@ -92,11 +107,28 @@ def _find_jvm_windows():
             jvm_dir = None
     return jvm_dir
 
+def _find_mac_lib(library):
+    jvm_dir = find_javahome()
+    for extension in (".dylib", ".jnilib"):
+        try:
+            cmd = ["find", os.path.dirname(jvm_dir), "-name", library+extension]
+            result = subprocess.check_output(cmd)
+            if type(result) == bytes:
+                lines = result.decode('utf-8').split("\n")
+            else:
+                lines = result.split("\n")
+            if len(lines) > 0 and len(lines[0]) > 0:
+                library_path = lines[0].strip()
+                return library_path
+        except Exception as e:
+            logger.error("Failed to execute \"%s\" when searching for %s" % 
+                         (cmd, library), exc_info=1)
+    logger.error("Failed to find %s (jvmdir: %s" % (library, jvm_dir))
+    return
+    
 def _find_jvm_mac():
-    # Put the jvm library on the path, hoping it is always in the same
-    # place.
-    jvm_dir = '/System/Library/Frameworks/JavaVM.framework/Libraries'
-    os.environ['PATH'] = os.environ['PATH'] + ':' + jvm_dir
+    # Load libjvm.dylib and lib/jli/libjli.dylib if it exists
+    jvm_dir = find_javahome()
     return jvm_dir
 
 def _find_jvm():
@@ -104,7 +136,7 @@ def _find_jvm():
     if sys.platform.startswith('win'):
         jvm_dir = _find_jvm_windows()
         if jvm_dir is None:
-            raise JavaNotFoundException()
+            raise JVMNotFoundError()
     elif sys.platform == 'darwin':
         jvm_dir = _find_jvm_mac()
     return jvm_dir
@@ -112,9 +144,22 @@ def _find_jvm():
 
 if sys.platform == "win32":
     # Need to fix up executable path to pick up jvm.dll
-    os.environ["PATH"] = os.environ["PATH"] + os.pathsep + _find_jvm()
-
-import _javabridge
+    # Also need path to JAVA_HOME\bin in order to find
+    # msvcr...
+    #
+    os.environ["PATH"] = os.environ["PATH"] + os.pathsep + _find_jvm() + \
+       os.pathsep + os.path.join(find_javahome(), "bin")
+    try:
+        os.add_dll_directory(_find_jvm())
+        os.add_dll_directory(os.path.join(find_javahome(), "bin"))
+    except AttributeError:
+        logger.debug("DLL directories not added to environment, may cause problems in Python 3.8+")
+    
+elif sys.platform == "darwin":
+    # Has side-effect of preloading dylibs
+    _find_jvm_mac()
+    
+import javabridge._javabridge as _javabridge
 __dead_event = threading.Event()
 __kill = [False]
 __main_thread_closures = []
@@ -144,6 +189,17 @@ class AtExit(object):
         
 __start_thread = None        
 
+class vm():
+    def __init__(self, *args, **kwds):
+        self.args = args
+        self.kwds = kwds
+
+    def __enter__(self):
+        start_vm(*self.args, **self.kwds)
+
+    def __exit__(self, type, value, traceback):
+        kill_vm()
+
 def start_vm(args=None, class_path=None, max_heap_size=None, run_headless=False):
     '''Start the Java Virtual Machine.
 
@@ -219,7 +275,11 @@ def start_vm(args=None, class_path=None, max_heap_size=None, run_headless=False)
         try:
             if sys.platform == "darwin":
                 logger.debug("Launching VM in non-python thread")
-                vm.create_mac(args, RQCLS)
+                library_path = _find_mac_lib("libjvm")
+                libjli_path = _find_mac_lib("libjli")
+                if library_path is None:
+                    raise Exception("Javabridge failed to find JVM library")
+                vm.create_mac(args, RQCLS, library_path, libjli_path)
                 logger.debug("Attaching to VM in monitor thread")
                 env = _javabridge.jb_attach()
             else:
@@ -315,7 +375,7 @@ def run_script(script, bindings_in = {}, bindings_out = {},
         scope = make_instance("org/mozilla/javascript/ImporterTopLevel",
                               "(Lorg/mozilla/javascript/Context;)V",
                               context)
-        for k, v in bindings_in.iteritems():
+        for k, v in bindings_in.items():
             call(scope, "put", 
                  "(Ljava/lang/String;Lorg/mozilla/javascript/Scriptable;"
                  "Ljava/lang/Object;)V", k, scope, v)
@@ -334,7 +394,7 @@ def run_script(script, bindings_in = {}, bindings_out = {},
                 "(Ljava/lang/String;"
                 "Lorg/mozilla/javascript/Scriptable;)"
                 "Ljava/lang/Object;", k, scope))
-    except JavaException, e:
+    except JavaException as e:
         if is_instance_of(e.throwable, "org/mozilla/javascript/WrappedException"):
             raise JavaException(call(e.throwable, "unwrap", "()Ljava/lang/Object;"))
         else:
@@ -608,7 +668,7 @@ def run_in_main_thread(closure, synchronous):
         def synchronous_closure():
             try:
                 result[0] = closure()
-            except Exception, e:
+            except Exception as e:
                 logger.exception("Caught exception when executing closure")
                 exception[0] = e
             done_event.set()
@@ -631,7 +691,7 @@ def print_all_stack_traces():
     for stak in stal:
         stakes = get_env().get_object_array_elements(stak)
         for stake in stakes:
-            print to_string(stake)
+            print(to_string(stake))
             
 CLOSE_ALL_WINDOWS = """
         new java.lang.Runnable() { 
@@ -639,8 +699,10 @@ CLOSE_ALL_WINDOWS = """
                 var all_frames = java.awt.Frame.getFrames();
                 if (all_frames) {
                     for (idx in all_frames) {
-                        java.lang.System.err.println("Disposing");
-                        all_frames[idx].dispose();
+                        try {
+                            all_frames[idx].dispose();
+                        } catch (err) {
+                        }
                     }
                 }
             }
@@ -1113,12 +1175,12 @@ def get_nice_arg(arg, sig):
         return make_instance('java/lang/Double', '(D)V', arg)
     if (sig in ('Ljava/lang/String;','Ljava/lang/Object;') and not
          isinstance(arg, _javabridge.JB_Object)):
-        if isinstance(arg, unicode):
-            arg, _ = codecs.utf_8_encode(arg)
-        elif arg is None:
+        if arg is None:
             return None
         else:
-            arg = str(arg)
+            if sys.version_info.major == 2:
+                if isinstance(arg, str):
+                    arg = arg.decode("utf-8")
         return env.new_string_utf(arg)
     if sig == 'Ljava/lang/Integer;' and type(arg) in [int, long, bool]:
         return make_instance('java/lang/Integer', '(I)V', int(arg))
@@ -1167,7 +1229,7 @@ def get_nice_result(result, sig):
     if (sig == 'Ljava/lang/String;' or
         (sig == 'Ljava/lang/Object;' and 
          is_instance_of(result, "java/lang/String"))):
-        return codecs.utf_8_decode(env.get_string_utf(result), 'replace')[0]
+        return env.get_string_utf(result)
     if sig == 'Ljava/lang/Integer;':
         return call(result, 'intValue', '()I')
     if sig == 'Ljava/lang/Long':
@@ -1270,7 +1332,7 @@ def get_collection_wrapper(collection, fn_wrapper=None):
     
         for d in get_collection_wrapper(list_of_hashmaps, get_map_wrapper):
             # a map wrapper on the hashmap is returned
-            print d["Foo"]
+            print(d["Foo"])
 
     '''
     class Collection(object):
@@ -1380,11 +1442,11 @@ def make_list(elements=[]):
 
     Examples::
         >>> mylist = make_list(["Hello", "World", 2])
-        >>> print "\\n".join([to_string(o) for o in mylist])
+        >>> print("\\n".join([to_string(o) for o in mylist]))
         Hello
         World
         2
-        >>> print "%s, %s." % (mylist[0], mylist[1].lower())
+        >>> print("%s, %s." % (mylist[0], mylist[1].lower()))
         Hello, world.
         >>> get_class_wrapper(mylist.o)
         java.util.ArrayList
@@ -1453,7 +1515,7 @@ def get_map_wrapper(o):
     
         > d = get_map_wrapper(jmap)
         > d["Foo"] = "Bar"
-        > print d["Foo"]
+        > print(d["Foo"])
         Bar
     '''
     assert is_instance_of(o, 'java/util/Map')
@@ -1486,6 +1548,9 @@ def get_map_wrapper(o):
         def __iter__(self):
             return iterate_collection(self.keySet())
         
+        def keys(self):
+            return tuple(iterate_collection(self.keySet(self)))
+        
     return Map()
 
 def make_map(**kwargs):
@@ -1494,7 +1559,7 @@ def make_map(**kwargs):
     Example::
     
         > d = make_map(foo="Bar")
-        > print d["foo"]
+        > print(d["foo"])
         Bar
         > get_class_wrapper(d.o)
         java.util.HashMap
@@ -1502,7 +1567,7 @@ def make_map(**kwargs):
         public java.lang.Object java.util.HashMap.put(java.lang.Object,java.lang.Object)
     '''
     hashmap = get_map_wrapper(make_instance('java/util/HashMap', "()V"))
-    for k, v in kwargs.iteritems():
+    for k, v in kwargs.items():
         hashmap[k] = v
     return hashmap
 
@@ -1540,7 +1605,7 @@ def get_enumeration_wrapper(enumeration):
     >>> enum = javabridge.get_enumeration_wrapper(keys)
     >>> while enum.hasMoreElements():
     ...     if javabridge.to_string(enum.nextElement()) == 'java.vm.name':
-    ...         print "Has java.vm.name"
+    ...         print("Has java.vm.name")
     ... 
     Has java.vm.name
 
@@ -1950,31 +2015,105 @@ def get_method_wrapper(obj):
         getModifiers = make_method('getModifiers', '()I')
     return Method()
         
-def attach_ext_env(env_address):
-    '''Attach to an externally supplied Java environment
-    
-    env_address - the numeric address of the env memory pointer
-    '''
-    global __thread_local_env 
-    env = _javabridge.JB_Env()
-    env.set_env(env_address)
-    __thread_local_env.env = env
-    
-def make_run_dictionary(jobject_address):
-    '''Support function for Py_RunString - jobject address -> globals / locals
+def make_run_dictionary(jobject):
+    '''Support function for Py_RunString - jobject -> globals / locals
     
     jobject_address - address of a Java Map of string to object
     '''
-    jmap = get_env().make_jb_object(jobject_address)
-    d = get_dictionary_wrapper(jmap)
-    
     result = {}
-    keys = jenumeration_to_string_list(d.keys())
-    for key in keys:
-        result[key] = d.get(key)
+    jmap = javabridge.JWrapper(jobject)
+    jentry_set = jmap.entrySet()
+    jentry_set_iterator = jentry_set.iterator()
+    while jentry_set_iterator.hasNext():
+        entry = jentry_set_iterator.next()
+        key, value = [o if not isinstance(o, javabridge.JWrapper) else o.o
+                      for o in (entry.getKey(), entry.getValue())]
+        result[to_string(key)] = value
     return result
 
+__weakrefdict = weakref.WeakValueDictionary()
+__strongrefdict = {}
 
+class _JRef(object):
+    '''A reference to some Python value for Java scripting
+    
+    A Java script executed using org.cellprofiler.javabridge.CPython.exec
+    might want to maintain and refer to objects and values. This class
+    wraps the value so that it can be referred to later.
+    '''
+    def __init__(self, value):
+        self.__value = value
+        
+    def __call__(self):
+        return self.__value
+    
+def create_jref(value):
+    '''Create a weak reference to a Python value
+    
+    This routine lets the Java method, CPython.exec(), create weak references
+    which can be redeemed by passing a token to redeem_weak_reference. The
+    routine returns a reference to the value as well and this reference must
+    be stored somewhere (e.g. a global value in a module) for the token to
+    be valid upon redemption.
+    
+    :param value: The value to be redeemed
+    
+    :returns: a tuple of a string token and a reference that must be maintained
+              in order to retrieve it later
+    '''
+    ref = _JRef(value)
+    ref_id = uuid.uuid4().hex
+    __weakrefdict[ref_id] = ref
+    return ref_id, ref
+
+def create_and_lock_jref(value):
+    '''Create and lock a value in one step
+    
+    :param value: the value to be redeemed
+    :returns: a ref_id that can be used to redeem the value and to unlock it.
+    '''
+    ref_id, ref = create_jref(value)
+    lock_jref(ref_id)
+    return ref_id
+
+def redeem_jref(ref_id):
+    '''Redeem a reference created using create_jref
+    
+    Raises KeyError if the reference could not be found, possibly because
+    someone didn't hold onto it
+    
+    :param ref_id: the token returned by create_jref for the reference
+    
+    :returns: the value
+    '''
+    return __weakrefdict[ref_id]()
+
+def lock_jref(ref_id):
+    '''Lock a reference to maintain it across CPython.exec() invocations
+    
+    Lock a reference into memory until unlock_jref is called. lock_jref()
+    can be called repeatedly on the same reference and the reference will
+    be held until an equal number of unlock_jref() calls have been made.
+    
+    :param ref_id: the ID returned from create_ref
+    '''
+    if ref_id not in __strongrefdict:
+        __strongrefdict[ref_id] = []
+    __strongrefdict[ref_id].append(__weakrefdict[ref_id])
+    
+def unlock_jref(ref_id):
+    '''Unlock a reference locked by lock_jref
+
+    Unlock and potentially dereference a reference locked by lock_jref()
+    
+    :param ref_id: the ID used to lock the reference
+    '''
+    refs = __strongrefdict[ref_id]
+    if len(refs) == 1:
+        del __strongrefdict[ref_id]
+    else:
+        refs.pop()
+        
 if __name__=="__main__":
     import wx
     app = wx.PySimpleApp(False)
